use std::time::Duration;
use std::{thread, time};
use serialport::{SerialPort, StopBits, Parity, FlowControl, DataBits};
use std::io::{self, Read, Write, BufWriter};
use regex::Regex;
use rustyline::Editor;
use rustyline::error::ReadlineError;
use std::fs::OpenOptions;
use std::io::{BufReader, BufRead};

pub mod perf {
    use chrono::{DateTime, Duration, Utc, Local};

    pub struct Perf {
        start: DateTime<Utc>,
        end: DateTime<Utc>,
        delta_ms: i64,
        delta_mks: i64,
    }

    impl Perf  {
        pub fn new() -> Perf {
            Perf {
                start: Utc::now(),
                end: Utc::now(),
                delta_ms: 0,
                delta_mks: 0,
            }
        }

        pub fn start(&mut self) {
            self.start = Utc::now();
        }

        // measure: "ms" or "mks"
        pub fn stop(&mut self, out_msg: &str, measure: &str) {
            self.end = Utc::now();
            self.delta_ms = (self.end - self.start).num_microseconds().unwrap()/1000;
            self.delta_mks = (self.end - self.start).num_microseconds().unwrap();

            if (measure == "ms") { println!("{} (ms): {:?}", out_msg, self.delta_ms); }
            else { println!("{} (mks): {:?}", out_msg, self.delta_mks); }
        }

        pub fn stop_restart(&mut self, out_msg: &str, measure: &str) {
            self.stop(out_msg, measure);
            self.start = Utc::now();
        }

        pub fn cur_time(format: &str) -> String {
            let now = Local::now();
            let fmt = now.format(format).to_string();
            return fmt;
        }

    }
}

fn read_str(s: &str) -> Result<Vec<String>, String> { 
    let re = match Regex::new(r#""([^"]+)"|(\S+)"#) {
        Ok(val) => val,
        Err(_) => { return Err(format!("Error in regex!"));  }
    };

    let res: Vec<String> = re.find_iter(&s)
        .map(|match_| match_.as_str().to_owned().replace("\"", ""))
        .collect();

    Ok(res)
}

fn set_port_settings(port_name: &str, baud_rate: u32, data_bits: DataBits, parity: Parity, 
                     stop_bits: StopBits, flow_ctrl: FlowControl, timeout_msec: u64) -> Result<Box<dyn SerialPort>, String> {

    let mut port = match serialport::new(port_name, baud_rate)
        .timeout(Duration::from_millis(timeout_msec)).open() {
            Ok(p) => p,
            Err(msg) => return Err(format!("{}", msg)),
        };

    port.set_baud_rate(baud_rate).unwrap();
    port.set_data_bits(data_bits).unwrap();
    port.set_parity(parity).unwrap();
    port.set_stop_bits(stop_bits).unwrap();
    port.set_flow_control(flow_ctrl).unwrap();
    return Ok(port);
}


fn tx_rx_data<W: std::io::Write>(port_name: &str, 
                                 mut data_out: String, 
                                 cmd_writer: &mut BufWriter<W>, 
                                 data_writer: &mut BufWriter<W>, 
                                 save_data_flag: bool,
                                 save_cmd_flag: bool) {
    use crate::perf::Perf;
    let mut perf = Perf::new();
    perf.start();
    // let port_name = &args[1];
    let mut port = match set_port_settings(port_name, 19200, 
                                           DataBits::Eight, 
                                           Parity::Even, 
                                           StopBits::One, 
                                           FlowControl::None, 100) {
        Ok(set) => {set},
        Err(msg) => { println!("error while opening com-port ({}): {}", port_name, msg); return }
    };

    perf.stop_restart("time open port", "ms");

    if save_cmd_flag == true {
        writeln!(cmd_writer, "{};{};{}", Perf::cur_time("%Y.%m.%d"), 
                 Perf::cur_time("%H:%M:%S.%f"), data_out.trim()).expect("err");
    }

    data_out.push_str(std::str::from_utf8(b"\x0D\x0A").unwrap()); // CR + LF
    match port.write(data_out.as_bytes()) {
        Ok(_) => {},
        Err(e) => eprintln!("Error sending data: {}", e),
    }

    port.flush().unwrap();

    perf.stop_restart("time write to port", "ms");

    let mut reader = BufReader::new(port);
    let mut s = String::new();
    reader.read_line(&mut s).unwrap();
    if s.len() > 0 {
        let data = format!("{}: {}", Perf::cur_time("%H:%M:%S.%f"), s.trim());
        println!("{}", data);
        if save_data_flag == true {
            writeln!(data_writer, "{}", 
                     format!("{};{};{}", 
                             Perf::cur_time("%Y.%m.%d"), 
                             Perf::cur_time("%H:%M:%S.%f"), 
                             s.trim()))
                .expect("err");
        }
    }

    perf.stop("time for receive data", "ms");

    // let mut buffer = Vec::new();
    // match port.read_to_end(&mut buffer) {
    //     Ok(_) => {
    //         if buffer.len() > 0 {
    //             let data = format!("{}: {:?}", cur_time("%H:%M:%S"), buffer);
    //             println!("{}", data);
    //             if save_data_flag == true {
    //                 writeln!(data_writer, "{}", 
    //                          format!("{};{};{}", 
    //                          cur_time("%Y.%m.%d"), 
    //                          cur_time("%H:%M:%S"), 
    //                          String::from_utf8(buffer).expect("msg")))
    //                     .expect("err");
    //             }
    //         }
    //     },
    //     Err(msg) => eprintln!("{}", msg),
    // }

    // let mut cnt: u32 = 0;
    // let mut i = 0;
    // let mut buf: String = String::new();

    // let start = Utc::now(); 
    // let mut timeout_flag = false;

    // let mut delta1 = 0;
    // while cnt == 0 {
    //     cnt = match port.bytes_to_read() {
    //         Ok(c) => c,
    //         Err(msg) => { println!("{}", msg); return; }
    //     };
    //     let delta = Utc::now() - start;
    //     delta1 =  delta.num_microseconds().unwrap()/1000;
    //     if delta1 > timeout_ms {
    //         println!("timeout: {}", delta1);
    //         timeout_flag = true;
    //         break;
    //     };
    // }

    // println!("time (ms): {}", delta1);

    // // println!("time (ms): {:?}", delta.num_microseconds().unwrap()/1000);

    // if timeout_flag == false {
    //     while i < 500000 {
    //         // thread::sleep(time::Duration::from_millis(10));
    //         i += 1;
    //         let mut buffer = [0u8; 4096];

    //         cnt = match port.bytes_to_read() {
    //             Ok(c) => c,
    //             Err(msg) => { println!("{}", msg); return; }
    //         };

    //         if cnt > 0 {
    //             match port.read(&mut buffer) {
    //                 Ok(count) => {
    //                     let s = std::str::from_utf8(&buffer[..count]).unwrap();
    //                     buf.push_str(s);
    //                     println!("{}", s);
    //                 },

    //                 Err(ref e) if e.kind() == io::ErrorKind::TimedOut => { println!("timeout!"); },
    //                 Err(e) => eprintln!("error reading from port ({}): {}", port_name, e),
    //             }
    //         }
    //         else {}
    //     }

    // }


    // if buf.len() > 0 {
    //     let data = format!("{}: {:?}", cur_time("%H:%M:%S"), buf);
    //     println!("{}", data);
    //     if save_data_flag == true {
    //         writeln!(data_writer, "{};{};{}", cur_time("%Y.%m.%d"), cur_time("%H:%M:%S"), buf.trim()).expect("err");
    //     }
    // }

}

fn main() {
    use crate::perf::Perf;
    let mut perf = Perf::new();

    let data_log_path = "../log/data.log";
    let file = OpenOptions::new()
        .write(true)
        .append(true)
        .open(data_log_path)
        .expect("error opening file ");

    let mut data_writer = BufWriter::new(file);

    let data_log_path = "../log/cmd.log";
    let file = OpenOptions::new()
        .write(true)
        .append(true)
        .open(data_log_path)
        .expect("error opening file ");

    let mut cmd_writer = BufWriter::new(file);

    let mut save_data_flag = false;
    let mut save_cmd_flag = false;

    let timeout_ms = 1000;
    let fname_hist = "../settings/cmd_history.log".to_string();
    let mut editor = Editor::<()>::new();
    if editor.load_history(&fname_hist).is_err() {
        println!("error while reading cmd_history.log file!", ); 
        return;
    } 

    loop {
        let s: String = match editor.readline(">> ") {
            Ok(s_) => s_.to_string(),
            Err(ReadlineError::Interrupted) => {
                editor.save_history(&fname_hist)
                    .expect("error while save the command to the history file (cmd_history.log)"); 
                return; 
            },
            Err(ReadlineError::Eof) => {
                editor.save_history(&fname_hist) 
                    .expect("error while save the command to the history file (cmd_history.log)"); 
                return; 
            },
            Err(msg) => {
                println!("{}{:?}",
                         "unknown command",
                         msg);
                return;
            }
        };

        editor.add_history_entry(&s);

        let args = match read_str(&s) {
            Ok(arg) => arg,
            Err(msg) => { 
                println!("{}", msg);
                continue;
            }
        };

        let args_cnt = args.len();

        let cmd_head = &args[0].clone();

        if cmd_head.contains("exit") { break; }
        else if cmd_head.contains("help") {
            println!("Serial port terminal supports the following commands:");
            println!("exit - excecute quit from this program");
            println!("help - explain how to use this program");
            println!("show_ports - shows all available serial ports");
            println!("send <port name> <data> - sends specified <data> to <port name>");
            println!("send_loop <port name> <data> <count times>  <period in msec> - sends specified <data> to <port name> in loop mode: <count times> times with period specified in <period in msec> ");
            println!("open_port <port name> <speed> <data bits> <parity> <stop bits> <flow control> - opens specified <port name> with specified settings");
            println!("save_data_log <enable/disable>");
            println!("save_cmd_log <enable/disable>");
        }
        else if cmd_head.contains("save_data_log") {
            if args_cnt != 2 { eprintln!("Error while parse command! {}", "arguments count must be 1"); }

            if args[1].contains("enable") { save_data_flag = true; }
            else { save_data_flag = false; }
        }
        else if cmd_head.contains("save_cmd_log") {
            if args_cnt != 2 { eprintln!("Error while parse command! {}", "arguments count must be 1"); }

            if args[1].contains("enable") { save_cmd_flag = true; }
            else { save_cmd_flag = false; }
        }
        else if cmd_head.contains("show_ports") {
            match serialport::available_ports() {
                Ok(ports) => {
                    for port in ports {
                        println!("{}: {:?}", port.port_name, port.port_type);
                    }
                }
                Err(msg) => { eprintln!("error while get all available serial port list: {}", msg); return; }
            }
        }
        else if cmd_head == "send" {
            if args_cnt != 3 { eprintln!("Error while parse command! {}", "arguments count must be 2"); }
            else {
                tx_rx_data(&args[1], 
                           args[2].clone(), 
                           &mut cmd_writer, 
                           &mut data_writer, 
                           save_data_flag,
                           save_cmd_flag);
            }
        }
        else if cmd_head == "send_loop" {
            if args_cnt != 5 { eprintln!("Error while parse command! {}", "arguments count must be 4"); }
            else {
                let cnt = args[3].parse::<u32>().unwrap();
                let period = args[4].parse::<u64>().unwrap();
                let mut i = 0;
                while (i < cnt) {
                    i += 1;
                    tx_rx_data(&args[1], 
                           args[2].clone(), 
                           &mut cmd_writer, 
                           &mut data_writer, 
                           save_data_flag,
                           save_cmd_flag);
                    
                    thread::sleep(time::Duration::from_millis(period));
                }
            }
        }
        else if cmd_head.contains("open port") {

            if args_cnt != 8 { eprintln!("Error while parse command! {}", "arguments count must be 8"); }
            else {
                // println!("open port <port name> <speed> <data bits> <parity> <stop bits> <flow control> - opens specified <port name> with specified settings");
                // let port_name = args[2].clone();
                // let speed = args[3].clone();
                // let data_bits = args[4].clone();
                // let parity = args[5].clone();
                // let stop_bits = args[6].clone();
                // let flow_ctrl = args[7].clone();
            }
        }
    }

    // close ports
    editor.save_history(&fname_hist).unwrap();
    // drop(port1);

}
